package sos.scheduler.file;


import sos.spooler.Job_impl;
import sos.spooler.Variable_set;
import sos.util.SOSLogger;
import sos.util.SOSSchedulerLogger;
import sos.util.SOSFileOperations;

/**
 * This job copies one file or several files of a directory.
 * It can be used standalone or as an order driven job.
 * 
 * @author Florian Schreiber <fs@sos-berlin.com>
 * @since  2006-11-01
*/
public class JobSchedulerCopyFile extends Job_impl {

	private SOSLogger logger = null;
	
    
	/**
	 * Implementierung fr Spooler API.
	 * Initialisierungmethode fr den Scheduler.
	 * @return boolean
	 */
	public boolean spooler_init() {
		
		try {
			try {
				this.logger = new SOSSchedulerLogger(this.spooler_log);
			} catch (Exception e) {
				throw new Exception("error occurred instantiating logger: " + e.getMessage());
			}
			
			return true;
			
		} catch (Exception e) {
			try {
				if ( logger!=null ) logger.error("error occurred in spooler_init(): " + e.getMessage());
			} catch (Exception x) {}
			
			return false;
		}
	}
	
    
	/**
	 * Implementierung fr Spooler API Luft bis return = false
	 * @return boolean
	 */
	public boolean spooler_process() {
		
		boolean rc = false;
		String name = null;
		
		String source = null;
		String target = null;
		String fileSpec = ".*";
		int flags = 0;
		String replacing = null;
		String replacement = null;
		
		try {
			
			// Job oder Order
			Variable_set params = spooler.create_variable_set();
			if (spooler_task.params() != null) params.merge(spooler_task.params());
			if (spooler_job.order_queue() != null && spooler_task.order().params() != null)
				params.merge(spooler_task.order().params());
			
			// mandatory parameters
			name = "source_file";
			if ( params.value(name)!=null && params.value(name).length()>0 )				
				source = params.value(name);
			else
				throw new Exception("job parameter is missing: ["+name+"]");
			logger.info(".. job parameter ["+name+"]: " + source);
			
			
			// optional parameters
			name = "target_file";
			if ( params.value(name)!=null && params.value(name).length()>0 ) {				
				target = params.value(name);
				logger.info(".. job parameter ["+name+"]: " + target);
			}
			
			name = "file_spec";
			if ( params.value(name)!=null && params.value(name).length()>0 ) {				
				fileSpec = params.value(name);
				logger.info(".. job parameter ["+name+"]: " + fileSpec);
			}
			
			try {
				
				name = "create_dir";
        if ( params.value(name) != null && params.value(name).length() > 0 ) {
					flags |= ( SOSFileOperations.toBoolean(params.value(name)) ) ? SOSFileOperations.CREATE_DIR : 0;
					logger.info(".. job parameter ["+name+"]: " + params.value(name));
				}
        
				name = "gracious";
        if ( params.value(name) != null && params.value(name).length() > 0 ) {
					flags |= ( SOSFileOperations.toBoolean(params.value(name)) ) ? SOSFileOperations.GRACIOUS : 0;
					logger.info(".. job parameter ["+name+"]: " + params.value(name));
				}
        
        name = "overwrite";
				if ( params.value(name)!=null && params.value(name).length()>0 ) {
					flags |= ( SOSFileOperations.toBoolean(params.value(name)) ) ? 0 : SOSFileOperations.NOT_OVERWRITE;
					logger.info(".. job parameter ["+name+"]: " + params.value(name));
				}
				
				name = "recursive";
				if ( params.value(name)!=null && params.value(name).length()>0 ) {
					flags |= ( SOSFileOperations.toBoolean(params.value(name)) ) ? SOSFileOperations.RECURSIVE : 0;
					logger.info(".. job parameter ["+name+"]: " + params.value(name));
				}
				
			} catch (Exception x) {
				throw new Exception("cannot evaluate job parameter ["+name+"]: " + x.getMessage());
			}
			
			name = "replacing";
			if ( params.value(name)!=null && params.value(name).length()>0 ) {				
				replacing = params.value(name);
				logger.info(".. job parameter ["+name+"]: " + replacing);
			}
			
			name = "replacement";
			// Es muss unterschieden werden zwischen NULL und ""
			String[] pArr = params.names().split(";");
			for (int i=0; i<pArr.length; i++) {
				if ( pArr[i].equals( name) )
					replacement = params.value(name);
			}
			
			if ( replacement!=null )
				logger.info(".. job parameter ["+name+"]: " + replacement);
			
			
			if ( replacing != null && replacement == null ) {
				throw new Exception("job parameter is missing for specified parameter [replacing]: [replacement]");
			}
      
			if ( replacing == null && replacement != null ) {
				throw new Exception("job parameter is missing for specified parameter [replacement]: [replacing]");
			}
			
			// Methodenaufruf
			if (replacing != null && replacement != null) {
				rc = SOSFileOperations.copyFile(source, target, fileSpec, flags, java.util.regex.Pattern.CASE_INSENSITIVE, replacing, replacement, logger);
			} else {
				rc = SOSFileOperations.copyFile(source, target, fileSpec, flags, java.util.regex.Pattern.CASE_INSENSITIVE, logger);
			}
			
			if (!rc && params.value("gracious") != null && params.value("gracious").equalsIgnoreCase("all")) {
				return (spooler_job.order_queue() != null);
			} else {
				return (spooler_job.order_queue() != null) ? rc : false;
			}
			
		} catch (Exception e) {
			
			try {
				logger.error("error occurred in JobSchedulerCopyFile: " + e.getMessage());
			} catch(Exception x) {}
			
			return false;
		}
	}
	
}
