package sos.scheduler.managed;

import java.io.ByteArrayInputStream;
import java.util.HashMap;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import sos.scheduler.job.JobSchedulerJob;
import sos.settings.SOSProfileSettings;
import sos.spooler.Order;
import sos.spooler.Variable_set;

/**
 * Base class for managed jobs. This class can be used to write own mangaged jobs.
 * By calling prepareParams() in spooler_process() task and order parameters
 * will be merged.
 *  
 * @author andreas.pueschel@sos-berlin.com
 * @since 1.0 2005-03-05 
 */
public abstract class JobSchedulerManagedJob extends JobSchedulerJob {

    
    /** Attribut: globalSettings: globale Einstellungen des Jobs */
    protected HashMap globalSettings 	= null;
    
    //  Payload des Orders oder Parameter des Jobs wenn kein Order
	protected sos.spooler.Variable_set orderPayload;
    
    //  luft der Job mit einem order?
	protected boolean orderJob = true;
        
    protected void prepareParams() throws Exception{
    	orderJob = !(spooler_task.job().order_queue() == null);
    	Order order = null;
    	if (orderJob == false) {    		
    		this.getLogger().debug3(spooler_job.name()+" running without order.");    		
    	}
    	try{
    		orderPayload = clearBlanks(spooler_task.params());        		
    		if(orderJob) {
    			order = spooler_task.order();        			
    			Variable_set orderPay = clearBlanks( order.params());
    			//getLogger().debug3("Cleared orderPayload: "+orderPay.xml());
    			orderPayload.merge(orderPay);
    			getLogger().debug6("Merged Payload: "+orderPayload.xml());
    		}
    	} catch(Exception e){
    		
    	}
    }
    
    private Variable_set clearBlanks(Variable_set set) throws Exception{
		Variable_set retSet = spooler.create_variable_set();		
		
		String[] keys = set.names().split(";");
		for(int i=0; i<keys.length; i++) {
			String parameterValue = set.var(keys[i]);
			if (parameterValue!=null && parameterValue.length()>0){
            	retSet.set_var(keys[i],parameterValue);
            }
		}
		
		return retSet;
	}
    
   
    
    protected void debugParamter(Variable_set params, String name){
    	try{
    		getLogger().debug6("Parameter: "+name+" value:\""+params.var(name)+"\"");
    	}catch (Exception e){}
    }
    
    /**
	 * @return Returns the orderJob.
	 */
	public boolean isOrderJob() {
		return orderJob;
	}
	/**
	 * @param orderJob The orderJob to set.
	 */
	public void setOrderJob(boolean orderJob) {
		this.orderJob = orderJob;
	}
	/**
	 * @return Returns the orderPayload.
	 */
	public sos.spooler.Variable_set getOrderPayload() {
		return orderPayload;
	}
}
