package sos.stresstest;


import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * @author Joacim Zschimmer <j@zsch.de>
 * @since 2004-08-10
 * @version 1.0
 *
 *Arguments berprft die Gltigkeit und Vollstndigkeit der Eingabeparameter
 */
public class Arguments
{
    static class Argument
    {
        Argument(String value )  { _value = value; }
        
        final String      _value;
        boolean           _read = false;
    };
    
    //-----------------------------------------------------------------------------------------
    
    final Map _arguments = new HashMap();

    //--------------------------------------------------------------------------------Arguments
    /**
     * Konstruktor, der die Eingabeparameter als Zeichenkettenfeld erhlt
     */
    Arguments( String[] args )
    throws
        Exception
    {
        for( int i = 0; i < args.length; i++ )
        {
            String arg = args[i];
            if( !arg.startsWith( "-" ) )  throw new Exception( "invalid parameter: " + arg );
     
            int equal = arg.indexOf( "=" );
            if( equal == -1 )  throw new Exception( "invalid parameter: " + arg );
            
            _arguments.put( arg.substring( 0, equal + 1 ), new Argument( arg.substring( equal + 1 ) ) );   
        }
    }

    //--------------------------------------------------------------------------------as_string
    
    String as_string( String option )
    throws
        Exception
    {
        return as_string( option, null );
    }
    
    //--------------------------------------------------------------------------------as_string
    
    String as_string( String option, String default_ )
    throws
        Exception
    {
        Argument argument = (Argument)_arguments.get( option );
        
        if( argument == null )
        {
            if( default_ == null )  throw new Exception( "parameter is missing: " + option );
            return default_;
        }
        
        argument._read = true;
        return argument._value;
    }

    //-----------------------------------------------------------------------------------as_int
    
    int as_int( String option )
    throws
        Exception
    {
        String result = as_string( option );
        
        try
        {
            return Integer.parseInt( result );
        }
        catch( Exception x )
        {
            throw new Exception( "parameter [" + option + "] is invalid: " + x, x );
        }
    }

    //-----------------------------------------------------------------------------------as_int
    
    int as_int( String option, int default_ )
    throws
        Exception
    {
        String result = as_string( option, "" );
        if( result.equals( "" ) )  return default_;
    
        return as_int( option );
    }
		
		//  -------------------------------------------------------------------------as_int_inRange
		
		int as_int_inRange ( String option, int min, int max, int default_ ) throws Exception {
			
			int value = as_int(option, default_);
			
			if ( value==default_ || (value>=min && value<=max) )
				return value;
			else
				throw new Exception( "parameter [" + option + "] not in range " + min + ".." + max );
		}
		
		//---------------------------------------------------------------------------check_all_used
		
    void check_all_used()
    throws
        Exception
    {
        for( Iterator it = _arguments.entrySet().iterator(); it.hasNext(); )
        {
            Map.Entry entry    = (Map.Entry)it.next();
            String    option   = (String)  entry.getKey();
            Argument  argument = (Argument)entry.getValue();
            
            if( !argument._read )  throw new Exception( "unknown parameter: " + option );
        }
    }
}

