// $Id: directory_observer.h 5430 2007-12-18 09:02:30Z jz $        Joacim Zschimmer, Zschimmer GmbH, http://www.zschimmer.com

#ifndef __SCHEDULER_DIRECTORY_OBSERVER_H
#define __SCHEDULER_DIRECTORY_OBSERVER_H

namespace sos {
namespace scheduler {
namespace directory_observer {

//-------------------------------------------------------------------------------------------------
    
struct Directory;

//-----------------------------------------------------------------------------------Directory_tree

struct Directory_tree : Scheduler_object,
                        Object
{
                                Directory_tree              ( Scheduler*, const file::File_path& directory_path );
                               ~Directory_tree              ();

    file::File_path             directory_path              () const                                { return _directory_path; }
    bool                        read                        ();
    void                        set_aging_until             ( double t )                            { if( _refresh_aged_entries_at > t )  _refresh_aged_entries_at = t; }
    void                        refresh_aged_entries        ();

  private:
    Fill_zero                  _zero_;
    file::File_path            _directory_path;
    ptr<Directory>             _root_directory;
    double                     _refresh_aged_entries_at;
};

//----------------------------------------------------------------------------------Directory_entry

struct Directory_entry
{
                                Directory_entry             ();

    Fill_zero                  _zero_;
    ptr<file::File_info>       _file_info;
    double                     _timestamp;        // Zeitstempel von _file_info
    ptr<Directory>             _subdirectory;               // ( _subdirectory != NULL ) == _file_info.is_directory()
    bool                       _is_aging;
    bool                       _is_removed;                 // _is_removed -> _is_aging
};

//----------------------------------------------------------------------------------------Directory

struct Directory : Object
{
                                Directory                   ( Directory_tree*, Directory* parent, const string& name );

    Absolute_path               path                        () const;
    bool                        read                        ();
    bool                        refresh_aged_entries        ();

  private:
    Fill_zero                  _zero_;
    Directory_tree* const      _directory_tree;
    Directory*                 _parent;
    string                     _name;
    list< Directory_entry >    _ordered_list;
};

//-------------------------------------------------------------------------------Directory_observer

struct Directory_observer : Scheduler_object,
                            Async_operation
{
                                Directory_observer          ( Scheduler*, const File_path& directory );
                               ~Directory_observer          ();

    void                        close                       ();



    // Async_operation
    bool                        async_finished_             () const                                { return false; }
    string                      async_state_text_           () const;
    bool                        async_continue_             ( Continue_flags );
    bool                        async_signaled_             ()                                      { return is_signaled(); }
    string                      obj_name                    () const                                { return Scheduler_object::obj_name(); }


    File_path                   directory_path              () const                                { return _directory_path; }

    bool                     is_signaled                    ()                                      { return _directory_event.signaled(); }
    void                    set_signaled                    ( const string& text )                  { _directory_event.set_signaled( text ); }

    void                        activate                    ();
    bool                        check                       ( double minimum_age = 0 );


  private:
    Fill_zero                  _zero_;
    file::File_path            _directory_path;
    Event                      _directory_event;
    ptr<Directory_tree>        _directory_tree;
    int                        _directory_watch_interval;
    double                     _last_change_at;
    double                     _read_again_at;
};

//-------------------------------------------------------------------------------------------------

} //namespace directory_observer

using namespace folder;

} //namespace scheduler
} //namespace sos

#endif
