// $Id: spooler_time.h 5417 2007-12-06 12:05:42Z jz $

#ifndef __SPOOLER_TIME_H
#define __SPOOLER_TIME_H

#include <math.h>

namespace sos {
namespace scheduler {
namespace time {

//-------------------------------------------------------------------------------------------------

extern const int                never_int;
const int                       max_weekdays_per_month      = 5;

//--------------------------------------------------------------------------------With_single_start

enum With_single_start
{
    wss_next_period                 = 0x01,                 // Nchste <period> mit end vor einem Zeitpunkt, begin ist egal
    wss_next_single_start           = 0x02,
    wss_next_period_or_single_start = 0x03,
    wss_next_begin                  = 0x04,                 // Nchste <period> mit begin ab einem Zeitpunkt
    wss_next_begin_or_single_start  = 0x06,
    wss_next_any_start              = 0x08                  // Nchster Start: single_start, start_once, repeat (nur erster Start in der Periode)
};

//---------------------------------------------------------------------------------------------Time

struct Time
{
    enum With_ms
    {
        without_ms,
        with_ms
    };


    enum Is_utc
    {
        is_utc
    };



    static Time                 time_with_now               ( const string& );              // Datum mit Zeit oder "now+zeit"
    static void             set_current_difference_to_utc   ( time_t now );
    static int                  current_difference_to_utc   ()                              { return static_current_difference_to_utc; }


                                Time                        ( double t = 0.0 )              { set(t); }
                                Time                        ( double t, Is_utc )            { set_utc( t ); }
                                Time                        ( time_t t )                    { set((double)t); }
                                Time                        ( time_t t, Is_utc )            { set_utc((double)t); }
                                Time                        ( int t )                       { set((double)t); }
                                Time                        ( uint t )                      { set((double)t); }
                                Time                        ( const string& t )             { set(t); }
                                Time                        ( const char* t   )             { set(t); }
                                Time                        ( const Sos_optional_date_time& dt ) { *this = dt; }
  //Z_WINDOWS_ONLY(             Time                        ( const FILETIME& t )           { set(t); } )
  //Z_WINDOWS_ONLY(             Time                        ( const SYSTEMTIME& t )         { set(t); } )

    void                        operator =                  ( double t )                    { set(t); }
    void                        operator =                  ( time_t t )                    { set((double)t); }
    void                        operator =                  ( int t )                       { set((double)t); }
    void                        operator =                  ( const string& t )             { set(t); }
    void                        operator =                  ( const char* t )               { set(t); }
    void                        operator =                  ( const Sos_optional_date_time& );

    void                        operator +=                 ( double t )                    { set( as_double() + t ); }
    void                        operator -=                 ( double t )                    { set( as_double() - t ); }

    Time                        operator +                  ( const Time& t )               { return Time( as_double() + t.as_double() ); }
    Time                        operator +                  ( double t )                    { return Time( as_double() + t             ); }
    Time                        operator +                  ( int t )                       { return Time( as_double() + t             ); }
    Time                        operator -                  ( const Time& t )               { return Time( as_double() - t.as_double() ); }
    Time                        operator -                  ( double t )                    { return Time( as_double() - t             ); }
    Time                        operator -                  ( int t )                       { return Time( as_double() - t             ); }

    bool                        operator <                  ( const Time& t ) const         { return as_double() <  t.as_double(); }
    bool                        operator <=                 ( const Time& t ) const         { return as_double() <= t.as_double(); }
    bool                        operator ==                 ( const Time& t ) const         { return as_double() == t.as_double(); }
    bool                        operator !=                 ( const Time& t ) const         { return as_double() != t.as_double(); }
    bool                        operator >=                 ( const Time& t ) const         { return as_double() >= t.as_double(); }
    bool                        operator >                  ( const Time& t ) const         { return as_double() >  t.as_double(); }

    bool                        operator <                  ( double t ) const              { return as_double() <  round(t); }
    bool                        operator <=                 ( double t ) const              { return as_double() <= round(t); }
    bool                        operator ==                 ( double t ) const              { return as_double() == round(t); }
    bool                        operator !=                 ( double t ) const              { return as_double() != round(t); }
    bool                        operator >=                 ( double t ) const              { return as_double() >= round(t); }
    bool                        operator >                  ( double t ) const              { return as_double() >  round(t); }

    bool                        operator <                  ( int t ) const                 { return as_double() <  round(t); }
    bool                        operator <=                 ( int t ) const                 { return as_double() <= round(t); }
    bool                        operator ==                 ( int t ) const                 { return as_double() == round(t); }
    bool                        operator !=                 ( int t ) const                 { return as_double() != round(t); }
    bool                        operator >=                 ( int t ) const                 { return as_double() >= round(t); }
    bool                        operator >                  ( int t ) const                 { return as_double() >  round(t); }

                                operator double             () const                        { return as_double(); }
    bool                        operator !                  () const                        { return is_null(); }

    void                    set_null                        ()                              { set( 0 ); }
    bool                     is_null                        () const                        { return _time == 0; }
    void                    set_never                       ()                              { set( never_int ); } 
    bool                     is_never                       () const                        { return _time == never_int; } 
    Time                        rounded_to_next_second      () const                        { return Time( floor( _time + 0.9995 ) ); }


    static double               round                       ( double t );
    static double               normalize                   ( double t );
    void                        set                         ( int t )                       { set( (double)t ); }
    void                        set                         ( time_t t )                    { set( (double)t ); }
    void                        set                         ( double );
    void                        set                         ( const string& );
    Time&                       set_utc                     ( double );
    double                      as_double                   () const;
    double                      as_utc_double               () const;

#ifdef Z_WINDOWS
    //void                        set                         ( const FILETIME& ); 
    //void                        set                         ( const SYSTEMTIME& );
    //FILETIME                    filetime                    () const;
#endif

    Time&                       set_datetime                ( const string& );
    void                        set_datetime_utc            ( const string& );
    Time                        time_of_day                 () const                        { return as_double() - midnight(); }
    Time                        midnight                    () const                        { return day_nr() * 24*60*60; }
    int                         day_nr                      () const                        { return uint(as_double()) / (24*60*60); }
    int                         month_nr                    () const;
    time_t                      as_time_t                   () const                        { return (time_t)( as_double    () + 0.0001 ); }
    time_t                      as_utc_time_t               () const                        { return (time_t)( as_utc_double() + 0.0001 ); }
    DATE                        as_local_com_date           () const                        { return com_date_from_seconds_since_1970( round( as_double() ) ); }
  //int64                       int64_filetime              () const;
    double                      cut_fraction                ( string* datetime_string );

    string                      as_string                   ( With_ms = with_ms ) const;                        
    string                      jdbc_value                  () const;
    string                      xml_value                   ( With_ms = with_ms ) const;                        
    void                        print                       ( ostream& s ) const            { s << as_string(); }
    friend ostream&             operator <<                 ( ostream& s, const Time& o )   { o.print(s); return s; }

    static Time                 now                         ();



#   if defined Z_DEBUG && defined Z_WINDOWS                 // Time in statischer Variablen fhrt mit gcc 3.3 zum Absturz in string::string
        string                 _time_as_string;
#   endif    

  private:
    double                     _time;                       // wie time_t: Anzahl Sekunden seit 1.1.1970 oder seit Mitternacht
    bool                       _is_utc;

  public:
    static int                  static_current_difference_to_utc;
    static const Time           never;
};      

void                            insert_into_message         ( Message_string*, int index, const Time& ) throw();


extern const Time               latter_day;

//-----------------------------------------------Daylight_saving_time_transition_detector_interface

struct Daylight_saving_time_transition_detector_interface : Async_operation
{
};


ptr<Daylight_saving_time_transition_detector_interface> new_daylight_saving_time_transition_detector( Scheduler* );

//-------------------------------------------------------------------------------------------Period

struct Period
{
                                Period                      ()                                      : _zero_(this+1) { init(); }
    explicit                    Period                      ( const xml::Element_ptr& e, const Period* deflt=NULL )  : _zero_(this+1) { init(); set_dom( e, without_date, deflt ); }
    
    void                        init                        ()                                      { _begin = _end = _repeat = _absolute_repeat = Time::never; }

    xml::Element_ptr            dom_element                 ( const xml::Document_ptr& ) const;

    bool                        empty                       () const                                { return _begin == Time::never; }
    bool                        has_start                   () const                                { return is_single_start() || repeat() != Time::never; }
    Time                        next_try                    ( const Time& );
    Period                      operator +                  ( const Time& t ) const                 { Period p = *this; p._begin += t; p._end += t; return p; }
    friend Period               operator +                  ( const Time& t, const Period& p )      { return p+t; }

    void                        set_default                 ();
    void                        set_single_start            ( const Time& );

    enum With_or_without_date { without_date, with_date };
    void                        set_dom                     ( const xml::Element_ptr&, With_or_without_date = without_date, const Period* deflt = NULL );

    bool                        operator <                  ( const Period& t ) const               { return _begin < t._begin; }  //fr set<>
    bool                        operator >                  ( const Period& t ) const               { return _begin > t._begin; }  //fr set<>
    bool                        is_in_time                  ( const Time& t )                       { return t >= _begin && t < _end; }
    bool                        is_comming                  ( const Time& time_of_day, With_single_start single_start ) const;

    Time                        begin                       () const                                { return _begin; }
    Time                        end                         () const                                { return _end; }
    Time                        repeat                      () const                                { return _repeat; }
    Time                        absolute_repeat             () const                                { return _absolute_repeat; }
    bool                        is_single_start             () const                                { return _single_start; }
    bool                        let_run                     () const                                { return _let_run; }
    bool                        has_repeat_or_once          () const                                { return _repeat < Time::never || _start_once; }
    Time                        next_repeated               ( const Time& ) const;
    Time                        next_absolute_repeated      ( const Time&, int next ) const;
  //Time                        next_repeated_before_end    ( const Time& ) const;

  //void                        set_next_start_time         ( const Time& );

    void                        check                       ( With_or_without_date ) const;
    string                      to_xml                      () const;
    void                        print                       ( ostream& ) const;
    string                      obj_name                    () const;
    friend ostream&             operator <<                 ( ostream& s, const Period& o )         { o.print(s); return s; }


//private:
    Fill_zero                  _zero_;
    Time                       _begin;                      // Sekunden seit Mitternacht, manchmal auch mit Datum
    Time                       _end;                        // Sekunden seit Mitternacht, manchmal auch mit Datum
    Time                       _repeat;
    Time                       _absolute_repeat;
    bool                       _single_start;
    bool                       _let_run;                    // Task zuende laufen lassen, nicht bei _next_end_time beenden
    bool                       _start_once;                 // Fr Joacim Zschimmer
};

//extern Period                   empty_period;               // gcc 3.2.1: Nicht const, weil gcc diese Variable nicht initialisiert. Das macht spooler.cxx.

typedef set<Period>             Period_set;

//----------------------------------------------------------------------------------------------Day

struct Day : Object
{
                                Day                         ()                                      {}
                                Day                         ( const Period_set& t )                 { _period_set = t; }
                                Day                         ( const Period& t )                     { _period_set.insert( t ); }
                                Day                         ( const xml::Element_ptr& e, const Day* deflt, const Period* p )   { set_dom_periods( e, deflt, p ); }

    void                        set_dom_periods             ( const xml::Element_ptr&, const Day* deflt, const Period* );
    void                        set_default                 ();

                                operator bool               () const                                { return !_period_set.empty(); }

    bool                        has_time                    ( const Time& time_of_day );
    Period                      next_period                 ( const Time& time_of_day, With_single_start single_start ) const { return _period_set.empty()? Period(): next_period_(time_of_day,single_start); }
    Period                      next_period_                ( const Time& time_of_day, With_single_start single_start ) const;
    void                        add                         ( const Period& p )                     { _period_set.insert( p ); }       

    void                        print                       ( ostream& ) const;
    friend ostream&             operator <<                 ( ostream& s, const Day& o )            { o.print(s); return s; }


    Period_set                 _period_set;
};

//------------------------------------------------------------------------------------------Day_set

struct Day_set
{
                                Day_set                     ( int minimum, int maximum );
    explicit                    Day_set                     ( int minimum, int maximum, const xml::Element_ptr& );

    void                        set_dom                     ( const xml::Element_ptr&, const Day* = NULL, const Period* = NULL );
    int                         get_day_number              ( const string& );
    list<int>                   get_day_numbers             ( const string& );
    int                         get_weekday_number          ( const string& );
    list<int>                   get_weekday_numbers         ( const string& );

    bool                        is_filled                   () const                                { return _is_day_set_filled; }

    void                        print                       ( ostream& ) const;
    friend ostream&             operator <<                 ( ostream& s, const Day_set& o )        { o.print(s); return s; }


  protected:
    Fill_zero                  _zero_;
    int                        _minimum;
    int                        _maximum;
    vector< ptr<Day> >         _days;
    bool                       _is_day_set_filled;
};

//--------------------------------------------------------------------------------------Weekday_set

struct Weekday_set : Day_set
{
                                Weekday_set                 ()                                      : Day_set( 0, 6 ) {}
    explicit                    Weekday_set                 ( const xml::Element_ptr& e )           : Day_set( 0, 6, e ) {}

    void                        fill_with_default           ( const Day& );
    Period                      next_period                 ( const Time&, With_single_start single_start );

    void                        print                       ( ostream& s ) const                    { s << "Weekday_set("; Day_set::print(s); s << ")"; }
    friend ostream&             operator <<                 ( ostream& s, const Weekday_set& o )    { o.print(s); return s; }
};

//--------------------------------------------------------------------------------------Monthday_set

struct Monthday_set : Day_set
{
                                Monthday_set                ()                                      : Day_set( 1, 31    ), _zero_(this+1) {}
    explicit                    Monthday_set                ( const xml::Element_ptr& e )           : Day_set( 1, 31, e ), _zero_(this+1) {}

    void                        set_dom                     ( const xml::Element_ptr&, const Day* = NULL, const Period* = NULL );

    bool                        is_filled                   () const                                { return Day_set::is_filled() | 
                                                                                                            _month_weekdays._is_filled | _reverse_month_weekdays._is_filled ; }
    Period                      next_period                 ( const Time&, With_single_start single_start );

    void                        print                       ( ostream& s ) const                    { s << "Monthday_set("; Day_set::print(s); s << ")"; }
    friend ostream&             operator <<                 ( ostream& s, const Monthday_set& o )   { o.print(s); return s; }


    //struct Month_weekday
    //{
    //    int                    _weekday_number;
    //    int                    _which;                      // 1...4: 1. bis 4. Wochentag des Monats, -1...-4 letzer bis viertletzer Wochentag des Monats
    //};

    //typedef stdext::hash_set<Month_weekday> Month_weekday_set;
    //Month_weekday_set          _month_weekday_set;

  private:
    Fill_zero                  _zero_;

    struct Month_weekdays
    {
                                Month_weekdays              ()                                      : _days( 7 * max_weekdays_per_month ), _is_filled(false) {}
        ptr<Day>&               day                         ( int which, int weekday_number );

        vector< ptr<Day> >     _days;
        bool                   _is_filled;
    };

    Month_weekdays             _month_weekdays;
    Month_weekdays             _reverse_month_weekdays;
};

//--------------------------------------------------------------------------------------Ultimo_set

struct Ultimo_set : Day_set
{
                                Ultimo_set                  ()                                      : Day_set( 0, 30 ) {}
    explicit                    Ultimo_set                  ( const xml::Element_ptr& e )           : Day_set( 0, 30, e ) {}

    Period                      next_period                 ( const Time&, With_single_start single_start );

    void                        print                       ( ostream& s ) const                    { s << "Ultimo_set("; Day_set::print(s); s << ")"; }
    friend ostream&             operator <<                 ( ostream& s, const Ultimo_set& o )     { o.print(s); return s; }
};

//--------------------------------------------------------------------------------------------Month

struct Month : Object
{
                                Month                       ()                                      : _zero_(this+1) {}

    void                        set_dom                     ( const xml::Element_ptr& );
    void                        set_default                 ();
    bool                        is_filled                   () const                                { return _weekday_set .is_filled() ||
                                                                                                             _monthday_set.is_filled() ||
                                                                                                             _ultimo_set  .is_filled();   }
    Period                      next_period                 ( const Time&, With_single_start );

    Fill_zero                  _zero_;
  //int                        _month_index;
    Weekday_set                _weekday_set;
    Monthday_set               _monthday_set;
    Ultimo_set                 _ultimo_set;                 // 0: Letzter Tag, -1: Vorletzter Tag
};

//-----------------------------------------------------------------------------------------Holidays

struct Holidays
{
                                Holidays                    ( Spooler* spooler )                    : _spooler(spooler) {}

    void                        clear                       ()                                      { _set.clear(); }
    bool                        is_filled                   () const                                { return !_set.empty(); }
    void                        set_dom                     ( const xml::Element_ptr&, int include_nesting = 0 );
    void                        include                     ( time_t t )                            { _set.insert( t ); }
    bool                        is_included                 ( const Time& t )                       { return _set.find( t.midnight().as_time_t() ) != _set.end(); }

    Spooler*                   _spooler;
    typedef stdext::hash_set<time_t> Set;
    Set                        _set;
};

//--------------------------------------------------------------------------------------------Date

struct Date
{
    bool                        operator <                  ( const Date& date ) const              { return _day_nr < date._day_nr; }

    void                        print                       ( ostream& ) const;
    friend ostream&             operator <<                 ( ostream& s, const Date& o )           { o.print(s); return s; }

    uint                       _day_nr;
    Day                        _day;
};

//----------------------------------------------------------------------------------------Date_set

struct Date_set
{
    Period                      next_period                 ( const Time&, With_single_start single_start );

    bool                        is_filled                   () const                                { return !_date_set.empty(); }
    void                        print                       ( ostream& ) const;
    friend ostream&             operator <<                 ( ostream& s, const Date_set& o )       { o.print(s); return s; }

    set<Date>                  _date_set;
};

//------------------------------------------------------------------------------------------At_set

struct At_set
{
    Period                      next_period                 ( const Time&, With_single_start single_start );

    bool                        is_filled                   () const                                { return !_at_set.empty(); }
    void                        add                         ( const Time& at )                      { _at_set.insert( at ); }
    void                        print                       ( ostream& ) const;
    friend ostream&             operator <<                 ( ostream& s, const At_set& o )         { o.print(s); return s; }

    set<Time>                  _at_set;
};

//----------------------------------------------------------------------------------------Run_time

struct Run_time : idispatch_implementation< Run_time, spooler_com::Irun_time >, 
                  spooler_com::Ihas_java_class_name
{
  //enum Application { application_job, application_order };


    static Class_descriptor     class_descriptor;
    static const Com_method    _methods[];


                                Run_time                    ( Scheduler_object* host_object );


    // IDispatch

    STDMETHODIMP_(ULONG)        AddRef                      ()                                      { return Idispatch_implementation::AddRef(); }
    STDMETHODIMP_(ULONG)        Release                     ()                                      { return Idispatch_implementation::Release(); }
    STDMETHODIMP                QueryInterface              ( const IID&, void** );


    // Ihas_java_class_name

    STDMETHODIMP            get_Java_class_name             ( BSTR* result )                        { return String_to_bstr( const_java_class_name(), result ); }
    STDMETHODIMP_(char*)  const_java_class_name             ()                                      { return (char*)"sos.spooler.Run_time"; }
    STDMETHODIMP            put_Xml                         ( BSTR xml );



    void                        close                       ();

    bool                        operator ==                 ( const Run_time& );

    bool                        is_filled                   () const;
    void                    set_modified_event_handler      ( Modified_event_handler* m )           { _modified_event_handler = m; }

    void                    set_xml                         ( const string& );
  //string                      xml                         ()                                      { return _xml; }

    void                    set_dom                         ( const xml::Element_ptr& );  // Setzt nicht _xml!
    xml::Element_ptr            dom_element                 ( const xml::Document_ptr& ) const;
    xml::Document_ptr           dom_document                () const;
    void                        append_calendar_dom_elements( const xml::Element_ptr&, Show_calendar_options* );


    int                         month_index_by_name         ( const string& );
    list<int>                   month_indices_by_names      ( const string& );
    void                        check                       ();                              

    bool                        set                         () const                                { return _set; }

    void                        set_holidays                ( const Holidays& h )                   { _holidays = h; }
    void                        set_default                 ();

    bool                        once                        ()                                      { return _once; }
    void                    set_once                        ( bool b = true )                       { _once = b; }

    Period                      first_period                ()                                      { return first_period( Time::now() ); }
    Period                      first_period                ( const Time& );

    Period                      next_period                 ( With_single_start single_start = wss_next_period )      { return next_period( Time::now(), single_start ); }
    Period                      next_period                 ( const Time&, With_single_start single_start = wss_next_period );
    Period                      call_function               ( const Time& beginning_time );

    bool                        period_follows              ( const Time& time )                    { return next_period(time).begin() != Time::never; }
  //bool                        period_follows              ( const Time& time )                    { return next_period( time, wss_next_period_or_single_start ).begin() != Time::never; }

    Time                        next_single_start           ( const Time& );
    Time                        next_any_start              ( const Time& );

    void                        print                       ( ostream& ) const;
    friend ostream&             operator <<                 ( ostream& s, const Run_time& o )       { o.print(s); return s; }



  private:
    Time                        next_time                   ( const Time& );


    Fill_zero                  _zero_;
    Spooler*                   _spooler;
    Scheduler_object*          _host_object;
  //Application                _application;
    Modified_event_handler*    _modified_event_handler;
    bool                       _set;
    bool                       _once;
    At_set                     _at_set;
    Date_set                   _date_set;
    Weekday_set                _weekday_set;
    Monthday_set               _monthday_set;
    Ultimo_set                 _ultimo_set;                 // 0: Letzter Tag, -1: Vorletzter Tag
    vector< ptr<Month> >       _months;
    Holidays                   _holidays;
  //string                     _xml;
    xml::Document_ptr          _dom;
    string                     _start_time_function;
    bool                       _start_time_function_error;
    ptr<Prefix_log>            _log;
};

//-------------------------------------------------------------------------------------------------

Time                            time_from_string            ( const string& );
void                            test_summertime             ( const string& date_time );
xml::Element_ptr                new_calendar_dom_element    ( const xml::Document_ptr&, const Time& );

//-------------------------------------------------------------------------------------------------

} //namespace time

//-------------------------------------------------------------------------------------------Gmtime
/* Hbsche Klasse, bercksichtigt aber bei der Umrechnug die Sommerzeit nicht.
// Greenwich Time (UTC)

struct Gmtime
{
    enum With_ms
    {
        without_ms,
        with_ms
    };

    static Gmtime               now                         ()                                      { return Gmtime( double_from_gmtime() ); }
    static double               round                       ( double t );
    static double               normalize                   ( double t );


                                Gmtime                      ()                                      : _time(0) {}
    explicit                    Gmtime                      ( const time::Time& t )                 { set_local_time( t ); }
    explicit                    Gmtime                      ( double t )                            { set( t ); }

                              //operator double             () const                                { return _time; }

  //Gmtime&                     operator =                  ( const Time& t )                       { set_local_time( t );  return *this; }
  //Gmtime&                     operator =                  ( time_t t )                            { _time = (double)t; }

    void                        operator +=                 ( double t )                            { set( _time + t ); }
    void                        operator -=                 ( double t )                            { set( _time - t ); }

    Gmtime                      operator +                  ( double t ) const                      { return Gmtime( _time + t ); }
    Gmtime                      operator +                  ( int    t ) const                      { return Gmtime( _time + (double)t ); }
    Gmtime                      operator -                  ( double t ) const                      { return Gmtime( _time - t ); }
    Gmtime                      operator -                  ( int    t ) const                      { return Gmtime( _time - (double)t ); }

    Gmtime                      operator -                  ( const Gmtime& t ) const               { return Gmtime( _time - t._time ); }

    bool                        operator <                  ( const Gmtime& t ) const               { return _time <  t._time; }
    bool                        operator <=                 ( const Gmtime& t ) const               { return _time <= t._time; }
    bool                        operator ==                 ( const Gmtime& t ) const               { return _time == t._time; }
    bool                        operator !=                 ( const Gmtime& t ) const               { return _time != t._time; }
    bool                        operator >=                 ( const Gmtime& t ) const               { return _time >= t._time; }
    bool                        operator >                  ( const Gmtime& t ) const               { return _time >  t._time; }

    bool                        operator <                  ( int t ) const                         { return _time <  t; }
    bool                        operator <=                 ( int t ) const                         { return _time <= t; }
    bool                        operator ==                 ( int t ) const                         { return _time == t; }
    bool                        operator !=                 ( int t ) const                         { return _time != t; }
    bool                        operator >=                 ( int t ) const                         { return _time >= t; }
    bool                        operator >                  ( int t ) const                         { return _time >  t; }

    void                    set_local_time                  ( const time::Time& t );
    time::Time                  local_time                  () const;
    
    bool                        is_null                     () const                                { return _time == 0; }
    double                      as_double                   () const                                { return _time; }
    string                      as_string                   ( With_ms = with_ms ) const;
    string                      as_local_string             ( With_ms = with_ms ) const;

    void                        set_null                    ()                                      { set( 0.0 ); }
    void                        set_time_t                  ( time_t t )                            { set( (double)t ); }
    void                        set                         ( double t );

    Gmtime                      time_of_day                 () const                                { return Gmtime( _time - midnight().as_double() ); }
    Gmtime                      midnight                    () const                                { return Gmtime( day_nr() * 24*60*60 ); }
    int                         day_nr                      () const                                { return (int)( int64(_time) / (24*60*60) ); }
    time_t                      as_time_t                   () const                                { return (time_t)( _time + 0.0001 ); }

  private:
#   if defined Z_DEBUG && defined Z_WINDOWS                 // Time in statischer Variablen fhrt mit gcc 3.3 zum Absturz in string::string
        string                 _time_as_string;
#   endif    

    double                     _time;
};

extern const Gmtime             doomsday;
*/
//-------------------------------------------------------------------------------------------------

inline void insert_into_message( Message_string* m, int index, const time::Time& t ) throw()        { m->insert_string( index, t.as_string() ); }

//-------------------------------------------------------------------------------------------------

using time::Time;
using time::Period;
using time::Run_time;

} //namespace scheduler
} //namespace sos

#endif
