// $Id: z_io.h 5419 2007-12-06 14:15:42Z jz $        Joacim Zschimmer, Zschimmer GmbH, http://www.zschimmer.com

#ifndef __ZSCHIMMER_Z_IO_H
#define __ZSCHIMMER_Z_IO_H

#include "md5.h"

namespace zschimmer {
namespace io {

//-------------------------------------------------------------------------------------------------

//template< typename SIMPLE_TYPE >
//struct sequence< SIMPLE_TYPE >
//{
//                                sequence                    ( const SIMPLE_TYPE* p, size_t length ) : _ptr(p), _length( length ) {}
//
//    int                         length                      () const                                { return _length; }
//    SIMPLE_TYPE                 operator[]                  ( size_t i ) const                      { assert( i < _length );  return _ptr[ i ]; }
//    sequence                    sub_sequence                ( size_t start, size_t end ) const      { assert( start < _length );  assert( end < _length );  assert( start <= end );
//                                                                                                      return sequence( _ptr + start, end - start ); }
//    string                      to_string                   () const                                { return string( _ptr, _length ); }
//
//  private:
//    const char* const          _ptr;
//    size_t const               _length;
//};

//-------------------------------------------------------------------------------------------------

extern const Byte               latin1_charset_table        [ 256 ];

inline bool                     is_valid_latin1             ( char c )                              { return latin1_charset_table[ (Byte)c ] != 0; }

//------------------------------------------------------------------------------------Char_sequence

struct Char_sequence
{
                                Char_sequence               ( const char* s )                       : _ptr(s), _length( strlen( s ) ) {}
                                Char_sequence               ( const char* p, size_t length )        : _ptr(p), _length( length ) {}
                                Char_sequence               ( const string* s )                     : _ptr( s->data() ), _length( s->length() ) {}
                                Char_sequence               ( const string& s )                     : _ptr( s.data() ), _length( s.length() ) {}

    const char*                 ptr                         () const                                { return _ptr; }
    size_t                      length                      () const                                { return _length; }
    bool                        is_empty                    () const                                { return _length == 0; }
    char                        operator[]                  ( size_t i ) const                      { Z_DEBUG_ONLY( assert( i < _length ) );  return _ptr[ i ]; }
    Char_sequence               sub_sequence                ( size_t start, size_t end ) const      { Z_DEBUG_ONLY( assert( start < _length );  assert( end < _length );  assert( start <= end ); )
                                                                                                      return Char_sequence( _ptr + start, end - start ); }
    bool                        is_valid_latin1             () const;
    string                      to_string                   () const                                { return string( _ptr, _length ); }

    friend ostream&             operator <<                 ( ostream& s, const Char_sequence& seq ){ s.write( seq.ptr(), seq.length() ); return s; }

  private:
    const char* const          _ptr;
    size_t const               _length;
};

//struct Char_sequence : IUnknown
//{
//    virtual                    ~Char_sequence               ()                                      = 0;
//
//    virtual int                 length                      () const                                = 0;
//    virtual char                operator[]                  ( int ) const                           = 0; 
//    virtual Char_sequence       sub_sequence                ( int start, int end ) const            = 0;
//    virtual string              to_string                   () const                                = 0;
//};

//-------------------------------------------------------------------------------------------------

struct Byte_sequence
{
                                Byte_sequence               ( const Byte* p, size_t length )        : _ptr(p), _length( length ) {}
                                Byte_sequence               ( const string* s )                     : _ptr( (const Byte*)s->data() ), _length( s->length() ) {}
                                Byte_sequence               ( const string& s )                     : _ptr( (const Byte*)s.data() ), _length( s.length() ) {}

    const Byte*                 ptr                         () const                                { return _ptr; }
    const char*                 char_ptr                    () const                                { return (const char*)_ptr; }
    size_t                      length                      () const                                { return _length; }
    Byte                        operator[]                  ( size_t i ) const                      { assert( i < _length );  return _ptr[ i ]; }
    Byte_sequence               sub_sequence                ( size_t start, size_t end ) const      { assert( start < _length );  assert( end < _length );  assert( start <= end );
                                                                                                      return Byte_sequence( _ptr + start, end - start ); }
    string                      to_string                   () const                                { return string( (const char*)_ptr, _length ); }

  private:
    const Byte* const          _ptr;
    size_t const               _length;
};

//-----------------------------------------------------------------------------String_char_sequence

//struct String_char_sequence : Char_sequence
//{
//                                String_char_sequence        ( const string* s )                     : Char_se_string(s) {}
//
//    int                         length                      () const                                { return s->length(); }
//    char                        operator[]                  ( int ) const                           { return s->length(); }
//    Char_sequence               sub_sequence                ( int start, int end ) const            { return _strings->length(); }
//    string                      to_string                   () const                                { return *_string; }
//
//  private:
//    string                     _string;
//};

//----------------------------------------------------------------------------------------Closeable
    
struct Closeable //: IUnknown
{
    virtual                    ~Closeable                   ()                                      {}
    virtual void                close                       ()                                      = 0;
};

//----------------------------------------------------------------------------------------Flushable
    
struct Flushable //: IUnknown
{
    virtual                    ~Flushable                   ()                                      {}
    virtual void                flush                       ()                                      = 0;
};

//-------------------------------------------------------------------------------------------Writer

struct Writer : IUnknown, Closeable, Flushable
{
    virtual void                write                       ( char c );
    virtual void                write                       ( const Char_sequence& )                = 0;  

  //void                        write                       ( const string& s )                     { write( Char_sequence( s.data(), s.length() ) ); }
};

//-------------------------------------------------------------------------------------------Writer

struct Filter_writer : simple_iunknown_implementation< Writer >
{
                                Filter_writer               ( Writer* w )                           : _writer(w) { assert( w ); }

    void                        write                       ( const Char_sequence& s )              { if( !_writer )  throw_xc( Z_FUNCTION );  _writer->write( s ); }
    void                        flush                       ()                                      {}
    void                        close                       ()                                      { if( _writer ) flush(), _writer = NULL; }
    //void                        flush                       ()                                      { _writer->flush(); }
    //void                        close                       ()                                      { ptr<Writer> w = _writer; _writer = NULL; w->close(); }

  protected:
    Writer*                    _writer;
};

//------------------------------------------------------------------------------------Output_stream

struct Output_stream : IUnknown, Closeable, Flushable
{
    virtual void                write_bytes                 ( const Byte_sequence& )                = 0;

    void                        write_bytes                 ( const string& s )                     { write_bytes( Byte_sequence( (const Byte*)s.data(), s.length() ) ); }
};

//-------------------------------------------------------------------------------------Input_stream

struct Input_stream : IUnknown, Closeable
{
    virtual                    ~Input_stream                ()                                      {}
    virtual string              read_bytes                  ()                                      = 0;    // sollten wir string nutzen?
};

//-----------------------------------------------------------------------------Filter_output_stream

struct Filter_output_stream : simple_iunknown_implementation< Output_stream >
{
                                Filter_output_stream        ( Output_stream* os )                   : _output_stream(os) { assert( os ); }
    virtual                    ~Filter_output_stream        ()                                      {}

    void                        close                       ()                                      { _output_stream->close();  _output_stream = NULL; }
    void                        flush                       ()                                      { _output_stream->flush(); }
    void                        write_bytes                 ( const Byte_sequence& b )              { _output_stream->write_bytes( b ); }
    string                      obj_name                    () const;

  protected:
    Output_stream*             _output_stream;
};

//-----------------------------------------------------------------Transparent_output_stream_writer

struct Transparent_output_stream_writer : simple_iunknown_implementation< Writer >
{
                                Transparent_output_stream_writer( Output_stream* );

    void                        write                       ( const Char_sequence& );
    void                        flush                       ()                                      { _output_stream->flush(); }
    void                        close                       ()                                      { _output_stream->close(); }

  private:
    ptr<Output_stream>         _output_stream;
};

//----------------------------------------------------------------------------------Buffered_writer

struct Buffered_writer : Filter_writer
{
    static const size_t         default_buffer_byte_count;


                                Buffered_writer             ( Writer*, int default_buffer_byte_count );
                               ~Buffered_writer             ();

    void                        write                       ( const Char_sequence& );
    void                        flush                       ();

  private:
    string                     _buffer;
    size_t                     _buffer_size;
};

//------------------------------------------------------------------------------------String_writer

struct String_writer : simple_iunknown_implementation< Writer >
{
                                String_writer               ();
                               ~String_writer               ();

    void                        write                       ( const Char_sequence& cs )             { _string_stream.append( cs.ptr(), cs.length() ); }
    void                        flush                       ()                                      {}
    void                        close                       ()                                      {}
    string                      to_string                   ()                                      { return _string_stream.to_string(); }

  private:
    String_stream              _string_stream;
};

//-------------------------------------------------------------------------------------------------

inline string                   lcase_hex                   ( const Byte_sequence& seq )            { return zschimmer::lcase_hex( seq.ptr(), seq.length() ); }            
inline string                   lcase_hex                   ( const Char_sequence& seq )            { return zschimmer::lcase_hex( seq.ptr(), seq.length() ); }            
inline bool                     is_valid_latin1             ( const Char_sequence& seq )            { return seq.is_valid_latin1(); }

//-------------------------------------------------------------------------------------------------

} //namespace io

inline void                     insert_into_message         ( Message_string* m, int index, const io::Char_sequence& seq ) { m->insert( index, string( seq.ptr(), seq.length() ) ); }

} //namespace zschimmer

#endif
